@@ -1,3 +1,4 @@
+#include <assert.h>
 #include <stdlib.h> /* for exit */
 #include <string.h> /* for strlen */
 #include <stdio.h> /* for fprintf etc */
@@ -17,7 +18,6 @@ static int new_label(struct generator * g) {
 }
 
 static struct str * vars_newname(struct generator * g) {
-
     struct str * output;
     g->var_number++;
     output = str_new();
@@ -29,7 +29,6 @@ static struct str * vars_newname(struct generator * g) {
 /* Write routines for items from the syntax tree */
 
 static void write_varname(struct generator * g, struct name * p) {
-
     if (p->type != t_external) {
         /* Pascal identifiers are case-insensitive but Snowball identifiers
          * should be case-sensitive.  To address this, we encode the case of
@@ -55,11 +54,11 @@ static void write_varname(struct generator * g, struct name * p) {
          * program authors to avoid naming externals which only differ by
          * case.
          */
-        int i, len = SIZE(p->b);
+        int len = SIZE(p->s);
         int lower_pending = 0;
         write_char(g, "SBIrxg"[p->type]);
-        for (i = 0; i != len; ++i) {
-            int ch = p->b[i];
+        for (int i = 0; i != len; ++i) {
+            int ch = p->s[i];
             if (ch >= 'a' && ch <= 'z') {
                 ++lower_pending;
             } else if (ch >= 'A' && ch <= 'Z') {
@@ -73,13 +72,12 @@ static void write_varname(struct generator * g, struct name * p) {
 
         write_char(g, '_');
     }
-    write_b(g, p->b);
+    write_s(g, p->s);
 }
 
 static void write_literal_string(struct generator * g, symbol * p) {
-    int i;
     write_char(g, '\'');
-    for (i = 0; i < SIZE(p); i++) {
+    for (int i = 0; i < SIZE(p); i++) {
         int ch = p[i];
         if (ch == '\'') {
             write_string(g, "''");
@@ -96,13 +94,26 @@ static void write_literal_string(struct generator * g, symbol * p) {
 }
 
 static void write_margin(struct generator * g) {
-    int i;
-    for (i = 0; i < g->margin; i++) write_string(g, "    ");
+    for (int i = 0; i < g->margin; i++) write_string(g, "    ");
+}
+
+static void write_relop(struct generator * g, int relop) {
+    switch (relop) {
+        case c_eq: write_string(g, " = "); break;
+        case c_ne: write_string(g, " <> "); break;
+        case c_gt: write_string(g, " > "); break;
+        case c_ge: write_string(g, " >= "); break;
+        case c_lt: write_string(g, " < "); break;
+        case c_le: write_string(g, " <= "); break;
+        default:
+            fprintf(stderr, "Unexpected type #%d in generate_integer_test\n", relop);
+            exit(1);
+    }
 }
 
 /* Write a variable declaration. */
 static void write_declare(struct generator * g,
-                          char * declaration,
+                          const char * declaration,
                           struct node * p) {
     struct str * temp = g->outbuf;
     g->outbuf = g->declarations;
@@ -126,7 +137,7 @@ static void write_block_start(struct generator * g) {
     w(g, "~MBegin~+~N");
 }
 
-static void write_block_end(struct generator * g) {   /* block end */
+static void write_block_end(struct generator * g) {
     w(g, "~-~MEnd;~N");
 }
 
@@ -139,22 +150,17 @@ static void write_savecursor(struct generator * g, struct node * p,
     writef(g, "~M~B0 := ~S1FCursor;~N" , p);
 }
 
-static void restore_string(struct node * p, struct str * out, struct str * savevar) {
-    str_clear(out);
+static void append_restore_string(struct node * p, struct str * out, struct str * savevar) {
     str_append_string(out, "FCursor := ");
     if (p->mode != m_forward) str_append_string(out, "FLimit - ");
     str_append(out, savevar);
     str_append_string(out, ";");
 }
 
-static void write_restorecursor(struct generator * g, struct node * p,
-                                struct str * savevar) {
-    struct str * temp = str_new();
+static void write_restorecursor(struct generator * g, struct node * p, struct str * savevar) {
     write_margin(g);
-    restore_string(p, temp, savevar);
-    write_str(g, temp);
+    append_restore_string(p, g->outbuf, savevar);
     write_newline(g);
-    str_delete(temp);
 }
 
 static void write_inc_cursor(struct generator * g, struct node * p) {
@@ -192,18 +198,18 @@ static void write_failure(struct generator * g) {
     switch (g->failure_label) {
         case x_return:
             write_string(g, "Begin Result := False; Exit; End;");
+            g->unreachable = true;
             break;
         default:
             write_string(g, "goto lab");
             write_int(g, g->failure_label);
             write_string(g, ";");
+            g->unreachable = true;
     }
     write_newline(g);
-    g->unreachable = true;
 }
 
-static void write_failure_if(struct generator * g, char * s, struct node * p) {
-
+static void write_failure_if(struct generator * g, const char * s, struct node * p) {
     writef(g, "~MIf (", p);
     writef(g, s, p);
     writef(g, ") Then~N", p);
@@ -224,18 +230,18 @@ static void write_check_limit(struct generator * g, struct node * p) {
 
 /* Formatted write. */
 static void writef(struct generator * g, const char * input, struct node * p) {
+    (void)p;
     int i = 0;
-    int l = strlen(input);
 
-    while (i < l) {
+    while (input[i]) {
         int ch = input[i++];
         if (ch != '~') {
             write_char(g, ch);
             continue;
         }
-        switch (input[i++]) {
-            default: write_char(g, input[i - 1]); continue;
-            case 'C': write_comment(g, p); continue;
+        ch = input[i++];
+        switch (ch) {
+            case '~': write_char(g, '~'); continue;
             case 'f': write_block_start(g);
                       write_failure(g);
                       g->unreachable = false;
@@ -245,21 +251,62 @@ static void writef(struct generator * g, const char * input, struct node * p) {
             case 'N': write_newline(g); continue;
             case '{': write_block_start(g); continue;
             case '}': write_block_end(g); continue;
-            case 'S': write_string(g, g->S[input[i++] - '0']); continue;
-            case 'B': write_b(g, g->B[input[i++] - '0']); continue;
-            case 'I': write_int(g, g->I[input[i++] - '0']); continue;
+            case 'S': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->S) / sizeof(g->S[0]))) {
+                    printf("Invalid escape sequence ~%c%c in writef(g, \"%s\", p)\n",
+                           ch, input[i - 1], input);
+                    exit(1);
+                }
+                write_string(g, g->S[j]);
+                continue;
+            }
+            case 'B': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->B) / sizeof(g->B[0])))
+                    goto invalid_escape2;
+                write_s(g, g->B[j]);
+                continue;
+            }
+            case 'I': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->I) / sizeof(g->I[0])))
+                    goto invalid_escape2;
+                write_int(g, g->I[j]);
+                continue;
+            }
             case 'V':
-            case 'W': write_varname(g, g->V[input[i++] - '0']); continue;
-            case 'L': write_literal_string(g, g->L[input[i++] - '0']); continue;
+            case 'W': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->V) / sizeof(g->V[0])))
+                    goto invalid_escape2;
+                write_varname(g, g->V[j]);
+                continue;
+            }
+            case 'L': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->L) / sizeof(g->L[0])))
+                    goto invalid_escape2;
+                write_literal_string(g, g->L[j]);
+                continue;
+            }
             case '+': g->margin++; continue;
             case '-': g->margin--; continue;
             case 'n': write_string(g, g->options->name); continue;
+            default:
+                printf("Invalid escape sequence ~%c in writef(g, \"%s\", p)\n",
+                       ch, input);
+                exit(1);
+            invalid_escape2:
+                printf("Invalid escape sequence ~%c%c in writef(g, \"%s\", p)\n",
+                       ch, input[i - 1], input);
+                exit(1);
         }
     }
 }
 
 static void w(struct generator * g, const char * s) {
-    writef(g, s, 0);
+    writef(g, s, NULL);
 }
 
 static void generate_AE(struct generator * g, struct node * p) {
@@ -312,26 +359,33 @@ static void generate_bra(struct generator * g, struct node * p) {
 }
 
 static void generate_and(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
 
     write_comment(g, p);
 
-    if (keep_c) write_savecursor(g, p, savevar);
+    if (savevar) write_savecursor(g, p, savevar);
 
     p = p->left;
     while (p) {
         generate(g, p);
         if (g->unreachable) break;
-        if (keep_c && p->right != 0) write_restorecursor(g, p, savevar);
+        if (savevar && p->right != NULL) write_restorecursor(g, p, savevar);
         p = p->right;
     }
-    str_delete(savevar);
+
+    if (savevar) {
+        str_delete(savevar);
+    }
 }
 
 static void generate_or(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
 
     int a0 = g->failure_label;
     struct str * a1 = str_copy(g->failure_str);
@@ -342,13 +396,13 @@ static void generate_or(struct generator * g, struct node * p) {
     write_comment(g, p);
     wsetlab_begin(g);
 
-    if (keep_c) write_savecursor(g, p, savevar);
+    if (savevar) write_savecursor(g, p, savevar);
 
     p = p->left;
     str_clear(g->failure_str);
 
-    if (p == 0) {
-        /* p should never be 0 after an or: there should be at least two
+    if (p == NULL) {
+        /* p should never be NULL after an or: there should be at least two
          * sub nodes. */
         fprintf(stderr, "Error: \"or\" node without children nodes.");
         exit(1);
@@ -363,7 +417,7 @@ static void generate_or(struct generator * g, struct node * p) {
         }
         wsetlab_end(g, g->failure_label);
         g->unreachable = false;
-        if (keep_c) write_restorecursor(g, p, savevar);
+        if (savevar) write_restorecursor(g, p, savevar);
         p = p->right;
     }
 
@@ -376,7 +430,10 @@ static void generate_or(struct generator * g, struct node * p) {
     if (!end_unreachable) {
         g->unreachable = false;
     }
-    str_delete(savevar);
+
+    if (savevar) {
+        str_delete(savevar);
+    }
 }
 
 static void generate_backwards(struct generator * g, struct node * p) {
@@ -388,14 +445,16 @@ static void generate_backwards(struct generator * g, struct node * p) {
 
 
 static void generate_not(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
 
-    int a0 = g->failure_label, l;
+    int a0 = g->failure_label;
     struct str * a1 = str_copy(g->failure_str);
 
     write_comment(g, p);
-    if (keep_c) {
+    if (savevar) {
         write_block_start(g);
         write_savecursor(g, p, savevar);
     }
@@ -404,7 +463,7 @@ static void generate_not(struct generator * g, struct node * p) {
     str_clear(g->failure_str);
 
     wsetlab_begin(g);
-    l = g->failure_label;
+    int l = g->failure_label;
 
     generate(g, p->left);
 
@@ -417,29 +476,37 @@ static void generate_not(struct generator * g, struct node * p) {
     wsetlab_end(g, l);
     g->unreachable = false;
 
-    if (keep_c) write_restorecursor(g, p, savevar);
-    if (keep_c) write_block_end(g);
-    str_delete(savevar);
+    if (savevar) {
+        write_restorecursor(g, p, savevar);
+        write_block_end(g);
+        str_delete(savevar);
+    }
 }
 
 
 static void generate_try(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
-
-    write_comment(g, p);
-    if (keep_c) write_savecursor(g, p, savevar);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
 
     g->failure_label = new_label(g);
     str_clear(g->failure_str);
-    if (keep_c) restore_string(p, g->failure_str, savevar);
+
+    write_comment(g, p);
+    if (savevar) {
+        write_savecursor(g, p, savevar);
+        append_restore_string(p, g->failure_str, savevar);
+    }
 
     wsetlab_begin(g);
     generate(g, p->left);
     wsetlab_end(g, g->failure_label);
     g->unreachable = false;
 
-    str_delete(savevar);
+    if (savevar) {
+        str_delete(savevar);
+    }
 }
 
 static void generate_set(struct generator * g, struct node * p) {
@@ -463,30 +530,35 @@ static void generate_fail(struct generator * g, struct node * p) {
 /* generate_test() also implements 'reverse' */
 
 static void generate_test(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
 
     write_comment(g, p);
 
-    if (keep_c) {
+    if (savevar) {
         write_savecursor(g, p, savevar);
     }
 
     generate(g, p->left);
 
-    if (!g->unreachable) {
-        if (keep_c) {
+    if (savevar) {
+        if (!g->unreachable) {
             write_restorecursor(g, p, savevar);
         }
+        str_delete(savevar);
     }
-    str_delete(savevar);
 }
 
 static void generate_do(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = K_needed(g, p->left);
+    struct str * savevar = NULL;
+    if (K_needed(g, p->left)) {
+        savevar = vars_newname(g);
+    }
+
     write_comment(g, p);
-    if (keep_c) write_savecursor(g, p, savevar);
+    if (savevar) write_savecursor(g, p, savevar);
 
     if (p->left->type == c_call) {
         /* Optimise do <call> */
@@ -503,24 +575,51 @@ static void generate_do(struct generator * g, struct node * p) {
         g->unreachable = false;
     }
 
-    if (keep_c) write_restorecursor(g, p, savevar);
-    str_delete(savevar);
+    if (savevar) {
+        write_restorecursor(g, p, savevar);
+        str_delete(savevar);
+    }
+}
+
+static void generate_next(struct generator * g, struct node * p) {
+    write_comment(g, p);
+    write_check_limit(g, p);
+    write_inc_cursor(g, p);
+}
+
+static void generate_GO_grouping(struct generator * g, struct node * p, int is_goto, int complement) {
+    write_comment(g, p);
+
+    struct grouping * q = p->name->grouping;
+    g->S[0] = p->mode == m_forward ? "" : "Bk";
+    g->S[1] = complement ? "In" : "Out";
+    g->V[0] = p->name;
+    g->I[0] = q->smallest_ch;
+    g->I[1] = q->largest_ch;
+    write_failure_if(g, "Not (Go~S1Grouping~S0(~V0, ~I0, ~I1))", p);
+    if (!is_goto) {
+        write_string(g, p->mode == m_forward ? "Inc(FCursor);" : "Dec(FCursor);");
+    }
 }
 
 static void generate_GO(struct generator * g, struct node * p, int style) {
-    int end_unreachable = false;
-    struct str * savevar = vars_newname(g);
-    int keep_c = style == 1 || repeat_restore(g, p->left);
+    write_comment(g, p);
+
     int a0 = g->failure_label;
     struct str * a1 = str_copy(g->failure_str);
 
+    int end_unreachable = false;
+
     int golab = new_label(g);
 
-    write_comment(g, p);
     w(g, "~MWhile True Do~N");
     w(g, "~{");
 
-    if (keep_c) write_savecursor(g, p, savevar);
+    struct str * savevar = NULL;
+    if (style == 1 || repeat_restore(g, p->left)) {
+        savevar = vars_newname(g);
+        write_savecursor(g, p, savevar);
+    }
 
     g->failure_label = new_label(g);
     str_clear(g->failure_str);
@@ -539,7 +638,10 @@ static void generate_GO(struct generator * g, struct node * p, int style) {
     }
     g->unreachable = false;
     wsetlab_end(g, g->failure_label);
-    if (keep_c) write_restorecursor(g, p, savevar);
+    if (savevar) {
+        write_restorecursor(g, p, savevar);
+        str_delete(savevar);
+    }
 
     g->failure_label = a0;
     str_delete(g->failure_str);
@@ -550,7 +652,6 @@ static void generate_GO(struct generator * g, struct node * p, int style) {
 
     g->I[0] = golab;
     w(g, "~}lab~I0:~N");
-    str_delete(savevar);
     g->unreachable = end_unreachable;
 }
 
@@ -572,13 +673,15 @@ static void generate_loop(struct generator * g, struct node * p) {
 }
 
 static void generate_repeat_or_atleast(struct generator * g, struct node * p, struct str * loopvar) {
-    struct str * savevar = vars_newname(g);
-    int keep_c = repeat_restore(g, p->left);
     int replab = new_label(g);
     g->I[0] = replab;
     writef(g, "lab~I0:~N~MWhile True Do~N~{", p);
 
-    if (keep_c) write_savecursor(g, p, savevar);
+    struct str * savevar = NULL;
+    if (repeat_restore(g, p->left)) {
+        savevar = vars_newname(g);
+        write_savecursor(g, p, savevar);
+    }
 
     g->failure_label = new_label(g);
     str_clear(g->failure_str);
@@ -586,7 +689,7 @@ static void generate_repeat_or_atleast(struct generator * g, struct node * p, st
     generate(g, p->left);
 
     if (!g->unreachable) {
-        if (loopvar != 0) {
+        if (loopvar != NULL) {
             g->B[0] = str_data(loopvar);
             w(g, "~MDec(~B0);~N");
         }
@@ -598,10 +701,12 @@ static void generate_repeat_or_atleast(struct generator * g, struct node * p, st
     wsetlab_end(g, g->failure_label);
     g->unreachable = false;
 
-    if (keep_c) write_restorecursor(g, p, savevar);
+    if (savevar) {
+        write_restorecursor(g, p, savevar);
+        str_delete(savevar);
+    }
 
     w(g, "~MBreak;~N~}");
-    str_delete(savevar);
 }
 
 static void generate_repeat(struct generator * g, struct node * p) {
@@ -615,7 +720,6 @@ static void generate_atleast(struct generator * g, struct node * p) {
     write_comment(g, p);
     w(g, "~{");
     g->B[0] = str_data(loopvar);
-
     write_declare(g, "~B0 : Integer", p);
     w(g, "~M~B0 := ");
     generate_AE(g, p->AE);
@@ -683,8 +787,8 @@ static void generate_hop(struct generator * g, struct node * p) {
         write_failure_if(g, "(C ~S1) Or (C ~S2 FCursor)", p);
     }
     writef(g, "~MFCursor := C;~N", p);
-    g->temporary_used = true;
     writef(g, "~}", p);
+    g->temporary_used = true;
 }
 
 static void generate_delete(struct generator * g, struct node * p) {
@@ -692,13 +796,6 @@ static void generate_delete(struct generator * g, struct node * p) {
     writef(g, "~MSliceDel;~N", p);
 }
 
-
-static void generate_next(struct generator * g, struct node * p) {
-    write_comment(g, p);
-    write_check_limit(g, p);
-    write_inc_cursor(g, p);
-}
-
 static void generate_tolimit(struct generator * g, struct node * p) {
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "FLimit" : "FBkLimit";
@@ -738,7 +835,7 @@ static void generate_sliceto(struct generator * g, struct node * p) {
 
 static void generate_address(struct generator * g, struct node * p) {
     symbol * b = p->literalstring;
-    if (b != 0) {
+    if (b != NULL) {
         write_literal_string(g, b);
     } else {
         write_varname(g, p->name);
@@ -746,7 +843,6 @@ static void generate_address(struct generator * g, struct node * p) {
 }
 
 static void generate_insert(struct generator * g, struct node * p, int style) {
-
     int keep_c = style == c_attach;
     write_comment(g, p);
     if (p->mode == m_backward) keep_c = !keep_c;
@@ -786,7 +882,6 @@ static void generate_slicefrom(struct generator * g, struct node * p) {
 }
 
 static void generate_setlimit(struct generator * g, struct node * p) {
-    struct str * savevar = vars_newname(g);
     struct str * varname = vars_newname(g);
     write_comment(g, p);
     if (p->left && p->left->type == c_tomark) {
@@ -799,6 +894,7 @@ static void generate_setlimit(struct generator * g, struct node * p) {
          * restore c.
          */
         struct node * q = p->left;
+        write_comment(g, q);
         g->S[0] = q->mode == m_forward ? ">" : "<";
         w(g, "~MIf (FCursor ~S0 "); generate_AE(g, q->AE); w(g, ") Then~N");
         write_block_start(g);
@@ -815,7 +911,8 @@ static void generate_setlimit(struct generator * g, struct node * p) {
             w(g, "~M~B0 := FBkLimit;~N");
             w(g, "~MFBkLimit := ");
         }
-        generate_AE(g, q->AE); writef(g, ";~N", q);
+        generate_AE(g, q->AE);
+        writef(g, ";~N", q);
 
         if (p->mode == m_forward) {
             str_assign(g->failure_str, "FLimit := FLimit + ");
@@ -827,7 +924,9 @@ static void generate_setlimit(struct generator * g, struct node * p) {
             str_append_ch(g->failure_str, ';');
         }
     } else {
+        struct str * savevar = vars_newname(g);
         write_savecursor(g, p, savevar);
+
         generate(g, p->left);
 
         if (!g->unreachable) {
@@ -852,6 +951,7 @@ static void generate_setlimit(struct generator * g, struct node * p) {
                 str_append_ch(g->failure_str, ';');
             }
         }
+        str_delete(savevar);
     }
 
     if (!g->unreachable) {
@@ -863,17 +963,16 @@ static void generate_setlimit(struct generator * g, struct node * p) {
             write_newline(g);
         }
     }
-
     str_delete(varname);
-    str_delete(savevar);
 }
 
 /* dollar sets snowball up to operate on a string variable as if it were the
  * current string */
 static void generate_dollar(struct generator * g, struct node * p) {
+    write_comment(g, p);
+
     struct str * savevar = vars_newname(g);
     g->B[0] = str_data(savevar);
-    write_comment(g, p);
     g->V[0] = p->name;
 
     {
@@ -899,7 +998,6 @@ static void generate_dollar(struct generator * g, struct node * p) {
     write_declare(g, "~B0_Ket : Integer", p);
     writef(g, "~{"
               "~M~B0_Current := FCurrent;~N"
-              "{ ~M~B0_Current := Copy(FCurrent, 1, FLimit); }~N"
               "~M~B0_Cursor := FCursor;~N"
               "~M~B0_Limit := FLimit;~N"
               "~M~B0_BkLimit := FBkLimit;~N"
@@ -918,12 +1016,12 @@ static void generate_dollar(struct generator * g, struct node * p) {
     str_delete(savevar);
 }
 
-static void generate_integer_assign(struct generator * g, struct node * p, char * s) {
-
+static void generate_integer_assign(struct generator * g, struct node * p, const char * s) {
+    write_comment(g, p);
     g->V[0] = p->name;
     w(g, "~M~W0 := ");
 
-    if (s != 0) {
+    if (s != NULL) {
         g->S[0] = s;
         w(g, "~W0 ~S0 ");
     }
@@ -932,29 +1030,62 @@ static void generate_integer_assign(struct generator * g, struct node * p, char
     w(g, ";~N");
 }
 
-static void generate_integer_test(struct generator * g, struct node * p, char * s) {
-
-    w(g, "~MIf Not (");
+static void generate_integer_test(struct generator * g, struct node * p) {
+    write_comment(g, p);
+    int relop = p->type;
+    int optimise_to_return = (g->failure_label == x_return && p->right && p->right->type == c_functionend);
+    if (optimise_to_return) {
+        w(g, "~MResult := ");
+        p->right = NULL;
+    } else {
+        w(g, "~MIf ");
+        // We want the inverse of the snowball test here.
+        relop ^= 1;
+    }
     generate_AE(g, p->left);
-    write_char(g, ' ');
-    write_string(g, s);
-    write_char(g, ' ');
+    write_relop(g, relop);
     generate_AE(g, p->AE);
-    w(g, ") Then~N");
-    write_block_start(g);
-    write_failure(g);
-    write_block_end(g);
-    g->unreachable = false;
+    if (optimise_to_return) {
+        w(g, "~N");
+    } else {
+        w(g, " Then~N");
+        write_block_start(g);
+        write_failure(g);
+        write_block_end(g);
+        g->unreachable = false;
+    }
 }
 
 static void generate_call(struct generator * g, struct node * p) {
-
+    int signals = check_possible_signals_list(g, p->name->definition, c_define, 0);
     write_comment(g, p);
     g->V[0] = p->name;
-    write_failure_if(g, "Not ~V0", p);
+    if (g->failure_label == x_return) {
+        if (p->right && p->right->type == c_functionend) {
+            /* Tail call. */
+            writef(g, "~MResult = ~V0;~N", p);
+            return;
+        }
+        if (signals == 0) {
+            /* Always fails. */
+            writef(g, "~MBegin; Result = ~V0; Exit; End;~N", p);
+            return;
+        }
+    }
+    if (signals == 1) {
+        /* Always succeeds. */
+        writef(g, "~M~V0;~N", p);
+    } else if (signals == 0) {
+        /* Always fails. */
+        writef(g, "~M~V0;~N", p);
+        write_failure(g);
+    } else {
+        write_failure_if(g, "Not ~V0", p);
+    }
 }
 
 static void generate_grouping(struct generator * g, struct node * p, int complement) {
+    write_comment(g, p);
 
     struct grouping * q = p->name->grouping;
     g->S[0] = p->mode == m_forward ? "" : "Bk";
@@ -966,7 +1097,6 @@ static void generate_grouping(struct generator * g, struct node * p, int complem
 }
 
 static void generate_namedstring(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "" : "Bk";
     g->V[0] = p->name;
@@ -983,17 +1113,19 @@ static void generate_literalstring(struct generator * g, struct node * p) {
 }
 
 static void generate_define(struct generator * g, struct node * p) {
-    struct str *saved_output;
-    struct str *saved_declarations;
+    struct name * q = p->name;
+    if (q->type == t_routine && !q->used) return;
+
+    write_newline(g);
+    write_comment(g, p);
 
     /* Generate function header. */
-    g->V[0] = p->name;
-    w(g, "~N~MFunction T~n.~W0 : Boolean;~N");
-
-    /* Save output*/
-    saved_output = g->outbuf;
-    saved_declarations = g->declarations;
+    g->V[0] = q;
+    w(g, "~MFunction T~n.~W0 : Boolean;~N");
 
+    /* Save output. */
+    struct str *saved_output = g->outbuf;
+    struct str *saved_declarations = g->declarations;
     g->outbuf = str_new();
     g->declarations = str_new();
 
@@ -1006,9 +1138,15 @@ static void generate_define(struct generator * g, struct node * p) {
 
     /* Generate function body. */
     w(g, "~{");
+    int signals = check_possible_signals_list(g, p->left, c_define, 0);
     g->temporary_used = false;
     generate(g, p->left);
-    if (!g->unreachable) w(g, "~MResult := True;~N");
+    if (p->left->right) {
+        assert(p->left->right->type == c_functionend);
+        if (signals) {
+            generate(g, p->left->right);
+        }
+    }
     w(g, "~}");
 
     if (g->temporary_used) {
@@ -1025,11 +1163,10 @@ static void generate_define(struct generator * g, struct node * p) {
     }
 
     if (g->next_label) {
-        int i, num = g->next_label;
-
         str_append_string(saved_output, "Label\n");
 
-        for (i = 0; i < num; ++i) {
+        int num = g->next_label;
+        for (int i = 0; i < num; ++i) {
             str_append_string(saved_output, "    lab");
             str_append_int(saved_output, i);
             str_append_string(saved_output, i == num - 1 ? ";\n" : ",\n");
@@ -1043,39 +1180,47 @@ static void generate_define(struct generator * g, struct node * p) {
     g->outbuf = saved_output;
 }
 
-static void generate_substring(struct generator * g, struct node * p) {
-    struct among * x = p->among;
+static void generate_functionend(struct generator * g, struct node * p) {
+    (void)p;
+    w(g, "~MResult := True;~N");
+}
 
+static void generate_substring(struct generator * g, struct node * p) {
     write_comment(g, p);
 
+    struct among * x = p->among;
+
     g->S[0] = p->mode == m_forward ? "" : "Bk";
     g->I[0] = x->number;
     g->I[1] = x->literalstring_count;
 
-    if (!x->amongvar_needed) {
-        write_failure_if(g, "FindAmong~S0(a_~I0, ~I1) = 0", p);
-    } else {
+    if (x->amongvar_needed) {
         writef(g, "~MAmongVar := FindAmong~S0(a_~I0, ~I1);~N", p);
-        write_failure_if(g, "AmongVar = 0", p);
+        if (!x->always_matches) {
+            write_failure_if(g, "AmongVar = 0", p);
+        }
+    } else if (x->always_matches) {
+        writef(g, "~MFindAmong~S0(a_~I0, ~I1);~N", p);
+    } else {
+        write_failure_if(g, "FindAmong~S0(a_~I0, ~I1) = 0", p);
     }
 }
 
 static void generate_among(struct generator * g, struct node * p) {
-
     struct among * x = p->among;
 
-    if (x->substring == 0) generate_substring(g, p);
-
-    if (x->starter != 0) generate(g, x->starter);
+    if (x->substring == NULL) {
+        generate_substring(g, p);
+    } else {
+        write_comment(g, p);
+    }
 
     if (x->command_count == 1 && x->nocommand_count == 0) {
         /* Only one outcome ("no match" already handled). */
         generate(g, x->commands[0]);
     } else if (x->command_count > 0) {
-        int i;
-        write_comment(g, p);
         w(g, "~MCase AmongVar Of~N~+");
-        for (i = 1; i <= x->command_count; i++) {
+        for (int i = 1; i <= x->command_count; i++) {
             g->I[0] = i;
             w(g, "~M~I0:~N~{");
             generate(g, x->commands[i - 1]);
@@ -1087,20 +1232,17 @@ static void generate_among(struct generator * g, struct node * p) {
 }
 
 static void generate_booltest(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     write_failure_if(g, "Not (~V0)", p);
 }
 
 static void generate_false(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     write_failure(g);
 }
 
 static void generate_debug(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->I[0] = g->debug_count++;
     g->I[1] = p->line_number;
@@ -1108,14 +1250,10 @@ static void generate_debug(struct generator * g, struct node * p) {
 }
 
 static void generate(struct generator * g, struct node * p) {
-
-    int a0;
-    struct str * a1;
-
     if (g->unreachable) return;
 
-    a0 = g->failure_label;
-    a1 = str_copy(g->failure_str);
+    int a0 = g->failure_label;
+    struct str * a1 = str_copy(g->failure_str);
 
     switch (p->type) {
         case c_define:        generate_define(g, p); break;
@@ -1133,6 +1271,11 @@ static void generate(struct generator * g, struct node * p) {
         case c_do:            generate_do(g, p); break;
         case c_goto:          generate_GO(g, p, 1); break;
         case c_gopast:        generate_GO(g, p, 0); break;
+        case c_goto_grouping: generate_GO_grouping(g, p, 1, 0); break;
+        case c_gopast_grouping:
+                              generate_GO_grouping(g, p, 0, 0); break;
+        case c_goto_non:      generate_GO_grouping(g, p, 1, 1); break;
+        case c_gopast_non:    generate_GO_grouping(g, p, 0, 1); break;
         case c_repeat:        generate_repeat(g, p); break;
         case c_loop:          generate_loop(g, p); break;
         case c_atleast:       generate_atleast(g, p); break;
@@ -1159,12 +1302,14 @@ static void generate(struct generator * g, struct node * p) {
         case c_minusassign:   generate_integer_assign(g, p, "-"); break;
         case c_multiplyassign:generate_integer_assign(g, p, "*"); break;
         case c_divideassign:  generate_integer_assign(g, p, "div"); break;
-        case c_eq:            generate_integer_test(g, p, "="); break;
-        case c_ne:            generate_integer_test(g, p, "<>"); break;
-        case c_gr:            generate_integer_test(g, p, ">"); break;
-        case c_ge:            generate_integer_test(g, p, ">="); break;
-        case c_ls:            generate_integer_test(g, p, "<"); break;
-        case c_le:            generate_integer_test(g, p, "<="); break;
+        case c_eq:
+        case c_ne:
+        case c_gt:
+        case c_ge:
+        case c_lt:
+        case c_le:
+            generate_integer_test(g, p);
+            break;
         case c_call:          generate_call(g, p); break;
         case c_grouping:      generate_grouping(g, p, false); break;
         case c_non:           generate_grouping(g, p, true); break;
@@ -1176,6 +1321,7 @@ static void generate(struct generator * g, struct node * p) {
         case c_false:         generate_false(g, p); break;
         case c_true:          break;
         case c_debug:         generate_debug(g, p); break;
+        case c_functionend:   generate_functionend(g, p); break;
         default: fprintf(stderr, "%d encountered\n", p->type);
                  exit(1);
     }
@@ -1213,54 +1359,65 @@ static void generate_method_decl(struct generator * g, struct name * q) {
 }
 
 static void generate_method_decls(struct generator * g) {
-    struct name * q;
-
     w(g, "~Mpublic~N~+");
     w(g, "~MConstructor Create;~N");
 
-    for (q = g->analyser->names; q; q = q->next) {
+    for (struct name * q = g->analyser->names; q; q = q->next) {
         if (q->type == t_external) {
             generate_method_decl(g, q);
         }
     }
     w(g, "~-");
 
-    w(g, "~Mprivate~N~+");
-    for (q = g->analyser->names; q; q = q->next) {
+    int first = true;
+    for (struct name * q = g->analyser->names; q; q = q->next) {
         if (q->type == t_routine) {
+            if (first) {
+                w(g, "~Mprivate~N~+");
+                first = false;
+            }
             generate_method_decl(g, q);
         }
     }
-    w(g, "~-");
+    if (!first) w(g, "~-");
 }
 
 static void generate_member_decls(struct generator * g) {
-    struct name * q;
-    w(g, "~Mprivate~N~+");
-    for (q = g->analyser->names; q; q = q->next) {
+    int first = true;
+    for (struct name * q = g->analyser->names; q; q = q->next) {
         g->V[0] = q;
         switch (q->type) {
             case t_string:
-                w(g, "~M~W0 : AnsiString;~N");
-                break;
             case t_integer:
-                w(g, "~M~W0 : Integer;~N");
-                break;
             case t_boolean:
-                w(g, "~M~W0 : Boolean;~N");
-                break;
+                if (first) {
+                    w(g, "~Mprivate~N~+");
+                    first = false;
+                }
+                switch (q->type) {
+                    case t_string:
+                        w(g, "~M~W0 : AnsiString;~N");
+                        break;
+                    case t_integer:
+                        w(g, "~M~W0 : Integer;~N");
+                        break;
+                    case t_boolean:
+                        w(g, "~M~W0 : Boolean;~N");
+                        break;
+                }
         }
     }
 
-    w(g, "~-");
+    if (!first) w(g, "~-");
 }
 
 static void generate_among_decls(struct generator * g) {
     struct among *a = g->analyser->amongs;
+    if (a == NULL) return;
 
     w(g, "~Mprivate~N~+");
 
-    while (a != 0) {
+    while (a != NULL) {
         g->I[0] = a->number;
         w(g, "~Ma_~I0 : Array Of TAmong;~N");
         a = a->next;
@@ -1270,33 +1427,33 @@ static void generate_among_decls(struct generator * g) {
 }
 
 static void generate_among_table(struct generator * g, struct among * x) {
-    int i;
+    write_comment(g, x->node);
+
     struct amongvec * v = x->b;
 
     g->I[0] = x->number;
     g->I[1] = x->literalstring_count;
 
     w(g, "~MSetLength(a_~I0, ~I1);~N~+");
-
-    for (i = 0; i < x->literalstring_count; i++, v++) {
+    for (int i = 0; i < x->literalstring_count; i++) {
         g->I[1] = i;
 
         /* Write among's string. */
-        g->L[0] = v->b;
+        g->L[0] = v[i].b;
         w(g, "~Ma_~I0[~I1].Str  := ~L0;~N");
 
         /* Write among's index & result. */
-        g->I[2] = v->i;
+        g->I[2] = v[i].i;
         w(g, "~Ma_~I0[~I1].Index := ~I2;~N");
-        g->I[2] = v->result;
+        g->I[2] = v[i].result;
         w(g, "~Ma_~I0[~I1].Result := ~I2;~N");
 
         /* Write among's handler. */
         w(g, "~Ma_~I0[~I1].Method := ");
-        if (v->function == 0) {
+        if (v[i].function == NULL) {
             w(g, "nil;~N~N");
         } else {
-            g->V[0] = v->function;
+            g->V[0] = v[i].function;
             w(g, "Self.~W0;~N~N");
         }
     }
@@ -1304,10 +1461,8 @@ static void generate_among_table(struct generator * g, struct among * x) {
 }
 
 static void generate_amongs(struct generator * g) {
-    struct among * a = g->analyser->amongs;
-    while (a != 0) {
-        generate_among_table(g, a);
-        a = a->next;
+    for (struct among * x = g->analyser->amongs; x; x = x->next) {
+        generate_among_table(g, x);
     }
 }
 
@@ -1319,8 +1474,9 @@ static void generate_constructor(struct generator * g) {
 
 static void generate_methods(struct generator * g) {
     struct node * p = g->analyser->program;
-    while (p != 0) {
+    while (p != NULL) {
         generate(g, p);
+        g->unreachable = false;
         p = p->right;
     }
 }
@@ -1328,22 +1484,19 @@ static void generate_methods(struct generator * g) {
 static void set_bit(symbol * b, int i) { b[i/8] |= 1 << i%8; }
 
 static void generate_grouping_table(struct generator * g, struct grouping * q) {
-
     int range = q->largest_ch - q->smallest_ch + 1;
     int size = (range + 7)/ 8;  /* assume 8 bits per symbol */
     symbol * b = q->b;
     symbol * map = create_b(size);
-    int i;
-    for (i = 0; i < size; i++) map[i] = 0;
 
-    /* Using unicode would require revision here */
+    for (int i = 0; i < size; i++) map[i] = 0;
 
-    for (i = 0; i < SIZE(b); i++) set_bit(map, b[i] - q->smallest_ch);
+    for (int i = 0; i < SIZE(b); i++) set_bit(map, b[i] - q->smallest_ch);
 
     g->V[0] = q->name;
     g->I[0] = size - 1;
     w(g, "~N~MConst~+~N~M~W0 : Array [0..~I0] Of Char = (~N~+");
-    for (i = 0; i < size; i++) {
+    for (int i = 0; i < size; i++) {
         if (i != 0) w(g, ",~N");
         g->I[0] = map[i];
         w(g, "~MChr(~I0)");
@@ -1354,8 +1507,7 @@ static void generate_grouping_table(struct generator * g, struct grouping * q) {
 }
 
 static void generate_groupings(struct generator * g) {
-    struct grouping * q;
-    for (q = g->analyser->groupings; q; q = q->next) {
+    for (struct grouping * q = g->analyser->groupings; q; q = q->next) {
         if (q->name->used)
             generate_grouping_table(g, q);
     }
