@@ -1,4 +1,4 @@
-
+#include <assert.h>
 #include <stdlib.h> /* for exit */
 #include <string.h> /* for strlen */
 #include <stdio.h> /* for fprintf etc */
@@ -17,7 +17,6 @@ static int new_label(struct generator * g) {
 }
 
 static struct str * vars_newname(struct generator * g) {
-
     struct str * output;
     g->var_number++;
     output = str_new();
@@ -26,11 +25,9 @@ static struct str * vars_newname(struct generator * g) {
     return output;
 }
 
-
 /* Write routines for items from the syntax tree */
 
 static void write_varname(struct generator * g, struct name * p) {
-
     switch (p->type) {
         case t_external:
             write_char(g, '_');
@@ -45,7 +42,7 @@ static void write_varname(struct generator * g, struct name * p) {
             break;
         }
     }
-    write_b(g, p->b);
+    write_s(g, p->s);
 }
 
 static void write_varref(struct generator * g, struct name * p) {
@@ -53,40 +50,43 @@ static void write_varref(struct generator * g, struct name * p) {
     write_varname(g, p);
 }
 
-static void write_hexdigit(struct generator * g, int n) {
-
-    write_char(g, n < 10 ? n + '0' : n - 10 + 'A');
-}
-
-static void write_hex(struct generator * g, int ch) {
-
-    write_string(g, "\\u");
-    {
-        int i;
-        for (i = 12; i >= 0; i -= 4) write_hexdigit(g, ch >> i & 0xf);
-    }
-}
-
 static void write_literal_string(struct generator * g, symbol * p) {
-
-    int i;
     write_string(g, "u\"");
-    for (i = 0; i < SIZE(p); i++) {
+    for (int i = 0; i < SIZE(p); i++) {
         int ch = p[i];
-        if (32 <= ch && ch < 127) {
-            if (ch == '\"' || ch == '\\') write_string(g, "\\");
-            write_char(g, ch);
+        if (32 <= ch && ch < 0x590 && ch != 127) {
+            if (ch == '"' || ch == '\\') write_char(g, '\\');
+            // Our Python generator uses ENC_WIDECHARS so we need to convert.
+            write_wchar_as_utf8(g, ch);
         } else {
-            write_hex(g, ch);
+            // Use escapes for anything over 0x590 as a crude way to avoid
+            // LTR characters affecting the rendering of source character
+            // order in confusing ways.
+            write_string(g, "\\u");
+            write_hex4(g, ch);
         }
     }
     write_string(g, "\"");
 }
 
-static void write_margin(struct generator * g) {
+static void write_literal_char(struct generator * g, symbol ch) {
+    write_string(g, "u\"");
+    if (32 <= ch && ch < 0x590 && ch != 127) {
+        if (ch == '"' || ch == '\\') write_char(g, '\\');
+        // Python uses ENC_WIDECHARS so we need to convert.
+        write_wchar_as_utf8(g, ch);
+    } else {
+        // Use escapes for anything over 0x590 as a crude way to avoid
+        // LTR characters affecting the rendering of source character
+        // order in confusing ways.
+        write_string(g, "\\u");
+        write_hex4(g, ch);
+    }
+    write_string(g, "\"");
+}
 
-    int i;
-    for (i = 0; i < g->margin; i++) write_string(g, "    ");
+static void write_margin(struct generator * g) {
+    for (int i = 0; i < g->margin; i++) write_string(g, "    ");
 }
 
 static void write_comment(struct generator * g, struct node * p) {
@@ -98,18 +98,15 @@ static void write_comment(struct generator * g, struct node * p) {
 }
 
 static void write_block_start(struct generator * g) {
-
     w(g, "~+~N");
 }
 
-static void write_block_end(struct generator * g)    /* block end */ {
-
+static void write_block_end(struct generator * g) {
     w(g, "~-");
 }
 
 static void write_savecursor(struct generator * g, struct node * p,
                              struct str * savevar) {
-
     g->B[0] = str_data(savevar);
     g->S[1] = "";
     if (p->mode != m_forward) g->S[1] = "self.limit - ";
@@ -117,7 +114,6 @@ static void write_savecursor(struct generator * g, struct node * p,
 }
 
 static void restore_string(struct node * p, struct str * out, struct str * savevar) {
-
     str_clear(out);
     str_append_string(out, "self.cursor = ");
     if (p->mode != m_forward) str_append_string(out, "self.limit - ");
@@ -126,7 +122,6 @@ static void restore_string(struct node * p, struct str * out, struct str * savev
 
 static void write_restorecursor(struct generator * g, struct node * p,
                                 struct str * savevar) {
-
     struct str * temp = str_new();
     write_margin(g);
     restore_string(p, temp, savevar);
@@ -136,14 +131,12 @@ static void write_restorecursor(struct generator * g, struct node * p,
 }
 
 static void write_inc_cursor(struct generator * g, struct node * p) {
-
     write_margin(g);
     write_string(g, p->mode == m_forward ? "self.cursor += 1" : "self.cursor -= 1");
     write_newline(g);
 }
 
 static void wsetlab_begin(struct generator * g) {
-
     w(g, "~Mtry:~N~+");
 }
 
@@ -158,7 +151,6 @@ static void wgotol(struct generator * g, int n) {
 }
 
 static void write_failure(struct generator * g) {
-
     if (str_len(g->failure_str) != 0) {
         write_margin(g);
         write_str(g, g->failure_str);
@@ -176,8 +168,7 @@ static void write_failure(struct generator * g) {
     }
 }
 
-static void write_failure_if(struct generator * g, char * s, struct node * p) {
-
+static void write_failure_if(struct generator * g, const char * s, struct node * p) {
     writef(g, "~Mif ", p);
     writef(g, s, p);
     writef(g, ":", p);
@@ -189,7 +180,6 @@ static void write_failure_if(struct generator * g, char * s, struct node * p) {
 
 /* if at limit fail */
 static void write_check_limit(struct generator * g, struct node * p) {
-
     if (p->mode == m_forward) {
         write_failure_if(g, "self.cursor >= self.limit", p);
     } else {
@@ -199,18 +189,18 @@ static void write_check_limit(struct generator * g, struct node * p) {
 
 /* Formatted write. */
 static void writef(struct generator * g, const char * input, struct node * p) {
+    (void)p;
     int i = 0;
-    int l = strlen(input);
 
-    while (i < l) {
+    while (input[i]) {
         int ch = input[i++];
         if (ch != '~') {
             write_char(g, ch);
             continue;
         }
-        switch (input[i++]) {
-            default: write_char(g, input[i - 1]); continue;
-            case 'C': write_comment(g, p); continue;
+        ch = input[i++];
+        switch (ch) {
+            case '~': write_char(g, '~'); continue;
             case 'f': write_block_start(g);
                       write_failure(g);
                       g->unreachable = false;
@@ -220,21 +210,65 @@ static void writef(struct generator * g, const char * input, struct node * p) {
             case 'N': write_newline(g); continue;
             case '{': write_block_start(g); continue;
             case '}': write_block_end(g); continue;
-            case 'S': write_string(g, g->S[input[i++] - '0']); continue;
-            case 'B': write_b(g, g->B[input[i++] - '0']); continue;
-            case 'I': write_int(g, g->I[input[i++] - '0']); continue;
-            case 'V': write_varref(g, g->V[input[i++] - '0']); continue;
-            case 'W': write_varname(g, g->V[input[i++] - '0']); continue;
-            case 'L': write_literal_string(g, g->L[input[i++] - '0']); continue;
+            case 'S': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->S) / sizeof(g->S[0]))) {
+                    printf("Invalid escape sequence ~%c%c in writef(g, \"%s\", p)\n",
+                           ch, input[i - 1], input);
+                    exit(1);
+                }
+                write_string(g, g->S[j]);
+                continue;
+            }
+            case 'B': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->B) / sizeof(g->B[0])))
+                    goto invalid_escape2;
+                write_s(g, g->B[j]);
+                continue;
+            }
+            case 'I': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->I) / sizeof(g->I[0])))
+                    goto invalid_escape2;
+                write_int(g, g->I[j]);
+                continue;
+            }
+            case 'V':
+            case 'W': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->V) / sizeof(g->V[0])))
+                    goto invalid_escape2;
+                if (ch == 'V')
+                    write_varref(g, g->V[j]);
+                else
+                    write_varname(g, g->V[j]);
+                continue;
+            }
+            case 'L': {
+                int j = input[i++] - '0';
+                if (j < 0 || j > (int)(sizeof(g->L) / sizeof(g->L[0])))
+                    goto invalid_escape2;
+                write_literal_string(g, g->L[j]);
+                continue;
+            }
             case '+': g->margin++; continue;
             case '-': g->margin--; continue;
             case 'n': write_string(g, g->options->name); continue;
+            default:
+                printf("Invalid escape sequence ~%c in writef(g, \"%s\", p)\n",
+                       ch, input);
+                exit(1);
+            invalid_escape2:
+                printf("Invalid escape sequence ~%c%c in writef(g, \"%s\", p)\n",
+                       ch, input[i - 1], input);
+                exit(1);
         }
     }
 }
 
 static void w(struct generator * g, const char * s) {
-    writef(g, s, 0);
+    writef(g, s, NULL);
 }
 
 static void generate_AE(struct generator * g, struct node * p) {
@@ -289,7 +323,6 @@ static void generate_AE(struct generator * g, struct node * p) {
 }
 
 static void generate_bra(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     p = p->left;
     while (p) {
@@ -299,7 +332,6 @@ static void generate_bra(struct generator * g, struct node * p) {
 }
 
 static void generate_and(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
 
@@ -311,14 +343,13 @@ static void generate_and(struct generator * g, struct node * p) {
     while (p) {
         generate(g, p);
         if (g->unreachable) break;
-        if (keep_c && p->right != 0) write_restorecursor(g, p, savevar);
+        if (keep_c && p->right != NULL) write_restorecursor(g, p, savevar);
         p = p->right;
     }
     str_delete(savevar);
 }
 
 static void generate_or(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
 
@@ -336,13 +367,13 @@ static void generate_or(struct generator * g, struct node * p) {
     p = p->left;
     str_clear(g->failure_str);
 
-    if (p == 0) {
-        /* p should never be 0 after an or: there should be at least two
+    if (p == NULL) {
+        /* p should never be NULL after an or: there should be at least two
          * sub nodes. */
         fprintf(stderr, "Error: \"or\" node without children nodes.");
         exit(1);
     }
-    while (p->right != 0) {
+    while (p->right != NULL) {
         int label = new_label(g);
         g->failure_label = label;
         wsetlab_begin(g);
@@ -370,7 +401,6 @@ static void generate_or(struct generator * g, struct node * p) {
 }
 
 static void generate_backwards(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     writef(g, "~Mself.limit_backward = self.cursor~N"
               "~Mself.cursor = self.limit~N", p);
@@ -380,7 +410,6 @@ static void generate_backwards(struct generator * g, struct node * p) {
 
 
 static void generate_not(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
 
@@ -414,7 +443,6 @@ static void generate_not(struct generator * g, struct node * p) {
 
 
 static void generate_try(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
     int label = new_label(g);
@@ -436,21 +464,18 @@ static void generate_try(struct generator * g, struct node * p) {
 }
 
 static void generate_set(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     writef(g, "~M~V0 = True~N", p);
 }
 
 static void generate_unset(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     writef(g, "~M~V0 = False~N", p);
 }
 
 static void generate_fail(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     generate(g, p->left);
     if (!g->unreachable) write_failure(g);
@@ -459,7 +484,6 @@ static void generate_fail(struct generator * g, struct node * p) {
 /* generate_test() also implements 'reverse' */
 
 static void generate_test(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
 
@@ -480,7 +504,6 @@ static void generate_test(struct generator * g, struct node * p) {
 }
 
 static void generate_do(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
     write_comment(g, p);
@@ -506,17 +529,19 @@ static void generate_do(struct generator * g, struct node * p) {
     str_delete(savevar);
 }
 
-static void generate_GO_grouping(struct generator * g, struct node * p, int is_goto, int complement) {
-
-    struct grouping * q = p->name->grouping;
+static void generate_next(struct generator * g, struct node * p) {
     write_comment(g, p);
+    write_check_limit(g, p);
+    write_inc_cursor(g, p);
+}
+
+static void generate_GO_grouping(struct generator * g, struct node * p, int is_goto, int complement) {
+    write_comment(g, p);
+
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->S[1] = complement ? "in" : "out";
-    g->S[2] = g->options->encoding == ENC_UTF8 ? "_U" : "";
     g->V[0] = p->name;
-    g->I[0] = q->smallest_ch;
-    g->I[1] = q->largest_ch;
-    write_failure_if(g, "not self.go_~S1_grouping~S0~S2(~n.~W0, ~I0, ~I1)", p);
+    write_failure_if(g, "not self.go_~S1_grouping~S0(~n.~W0)", p);
     if (!is_goto) {
         if (p->mode == m_forward)
             w(g, "~Mself.cursor += 1~N");
@@ -526,42 +551,21 @@ static void generate_GO_grouping(struct generator * g, struct node * p, int is_g
 }
 
 static void generate_GO(struct generator * g, struct node * p, int style) {
-    int end_unreachable;
-    struct str * savevar;
-    int keep_c;
-
-    int a0;
-    struct str * a1;
-
-    int golab;
-    int label;
-
-    if (p->left->type == c_grouping || p->left->type == c_non) {
-        /* Special case for "goto" or "gopast" when used on a grouping or an
-         * inverted grouping - the movement of c by the matching action is
-         * exactly what we want! */
-#ifdef OPTIMISATION_WARNINGS
-        printf("Optimising %s %s\n", style ? "goto" : "gopast", p->left->type == c_non ? "non" : "grouping");
-#endif
-        write_comment(g, p);
-        generate_GO_grouping(g, p->left, style, p->left->type == c_non);
-        return;
-    }
-
-    end_unreachable = false;
-    savevar = vars_newname(g);
-    keep_c = style == 1 || repeat_restore(g, p->left);
-
-    a0 = g->failure_label;
-    a1 = str_copy(g->failure_str);
-
-    golab = new_label(g);
     write_comment(g, p);
+
+    int end_unreachable = false;
+    struct str * savevar = vars_newname(g);
+    int keep_c = style == 1 || repeat_restore(g, p->left);
+
+    int a0 = g->failure_label;
+    struct str * a1 = str_copy(g->failure_str);
+
+    int golab = new_label(g);
     w(g, "~Mtry:~N~+"
              "~Mwhile True:~N~+");
     if (keep_c) write_savecursor(g, p, savevar);
 
-    label = new_label(g);
+    int label = new_label(g);
     g->failure_label = label;
     str_clear(g->failure_str);
     wsetlab_begin(g);
@@ -595,14 +599,24 @@ static void generate_GO(struct generator * g, struct node * p, int style) {
 }
 
 static void generate_loop(struct generator * g, struct node * p) {
-
     struct str * loopvar = vars_newname(g);
     write_comment(g, p);
     g->B[0] = str_data(loopvar);
-    w(g, "~Mfor ~B0 in range (");
-    generate_AE(g, p->AE);
-    g->B[0] = str_data(loopvar);
-    writef(g, ", 0, -1):~N", p);
+    if (p->AE->type == c_number && p->AE->number <= 4) {
+        // Use a tuple instead of range() for small constant numbers of
+        // iterations.
+        w(g, "~Mfor ~B0 in ");
+        for (int i = p->AE->number; i > 0; --i) {
+            w(g, "0");
+            if (i > 1) w(g, ", ");
+        }
+        writef(g, ":~N", p);
+    } else {
+        w(g, "~Mfor ~B0 in range(");
+        generate_AE(g, p->AE);
+        g->B[0] = str_data(loopvar);
+        writef(g, "):~N", p);
+    }
     writef(g, "~{", p);
 
     generate(g, p->left);
@@ -613,7 +627,6 @@ static void generate_loop(struct generator * g, struct node * p) {
 }
 
 static void generate_repeat_or_atleast(struct generator * g, struct node * p, struct str * loopvar) {
-
     struct str * savevar = vars_newname(g);
     int keep_c = repeat_restore(g, p->left);
     int label = new_label(g);
@@ -626,7 +639,7 @@ static void generate_repeat_or_atleast(struct generator * g, struct node * p, st
     generate(g, p->left);
 
     if (!g->unreachable) {
-        if (loopvar != 0) {
+        if (loopvar != NULL) {
             g->B[0] = str_data(loopvar);
             w(g, "~M~B0 -= 1~N");
         }
@@ -649,7 +662,6 @@ static void generate_repeat(struct generator * g, struct node * p) {
 }
 
 static void generate_atleast(struct generator * g, struct node * p) {
-
     struct str * loopvar = vars_newname(g);
     write_comment(g, p);
     g->B[0] = str_data(loopvar);
@@ -672,14 +684,12 @@ static void generate_atleast(struct generator * g, struct node * p) {
 }
 
 static void generate_setmark(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     writef(g, "~M~V0 = self.cursor~N", p);
 }
 
 static void generate_tomark(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? ">" : "<";
 
@@ -692,7 +702,6 @@ static void generate_tomark(struct generator * g, struct node * p) {
 }
 
 static void generate_atmark(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     w(g, "~Mif self.cursor != "); generate_AE(g, p->AE); writef(g, ":", p);
     write_block_start(g);
@@ -702,7 +711,6 @@ static void generate_atmark(struct generator * g, struct node * p) {
 }
 
 static void generate_hop(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "+" : "-";
 
@@ -725,30 +733,19 @@ static void generate_hop(struct generator * g, struct node * p) {
 }
 
 static void generate_delete(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     writef(g, "~Mif not self.slice_del():~N"
               "~+~Mreturn False~N~-"
               "~N", p);
 }
 
-
-static void generate_next(struct generator * g, struct node * p) {
-
-    write_comment(g, p);
-    write_check_limit(g, p);
-    write_inc_cursor(g, p);
-}
-
 static void generate_tolimit(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "self.limit" : "self.limit_backward";
     writef(g, "~Mself.cursor = ~S0~N", p);
 }
 
 static void generate_atlimit(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "self.limit" : "self.limit_backward";
     g->S[1] = p->mode == m_forward ? "<" : ">";
@@ -756,28 +753,24 @@ static void generate_atlimit(struct generator * g, struct node * p) {
 }
 
 static void generate_leftslice(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "self.bra" : "self.ket";
     writef(g, "~M~S0 = self.cursor~N", p);
 }
 
 static void generate_rightslice(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "self.ket" : "self.bra";
     writef(g, "~M~S0 = self.cursor~N", p);
 }
 
 static void generate_assignto(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     writef(g, "~M~V0 = self.assign_to()~N", p);
 }
 
 static void generate_sliceto(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     writef(g, "~M~V0 = self.slice_to()~N"
@@ -786,9 +779,8 @@ static void generate_sliceto(struct generator * g, struct node * p) {
 }
 
 static void generate_address(struct generator * g, struct node * p) {
-
     symbol * b = p->literalstring;
-    if (b != 0) {
+    if (b != NULL) {
         write_literal_string(g, b);
     } else {
         write_varref(g, p->name);
@@ -796,7 +788,6 @@ static void generate_address(struct generator * g, struct node * p) {
 }
 
 static void generate_insert(struct generator * g, struct node * p, int style) {
-
     int keep_c = style == c_attach;
     write_comment(g, p);
     if (p->mode == m_backward) keep_c = !keep_c;
@@ -808,7 +799,6 @@ static void generate_insert(struct generator * g, struct node * p, int style) {
 }
 
 static void generate_assignfrom(struct generator * g, struct node * p) {
-
     int keep_c = p->mode == m_forward; /* like 'attach' */
 
     write_comment(g, p);
@@ -823,9 +813,7 @@ static void generate_assignfrom(struct generator * g, struct node * p) {
     if (keep_c) w(g, "~Mself.cursor = c~N");
 }
 
-
 static void generate_slicefrom(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     w(g, "~Mif not self.slice_from(");
     generate_address(g, p);
@@ -847,6 +835,7 @@ static void generate_setlimit(struct generator * g, struct node * p) {
          * restore c.
          */
         struct node * q = p->left;
+        write_comment(g, q);
         g->S[0] = q->mode == m_forward ? ">" : "<";
         w(g, "~Mif self.cursor ~S0 "); generate_AE(g, q->AE); w(g, ":");
         write_block_start(g);
@@ -913,7 +902,6 @@ static void generate_setlimit(struct generator * g, struct node * p) {
 /* dollar sets snowball up to operate on a string variable as if it were the
  * current string */
 static void generate_dollar(struct generator * g, struct node * p) {
-
     struct str * savevar = vars_newname(g);
     g->B[0] = str_data(savevar);
     write_comment(g, p);
@@ -944,48 +932,72 @@ static void generate_dollar(struct generator * g, struct node * p) {
     str_delete(savevar);
 }
 
-static void generate_integer_assign(struct generator * g, struct node * p, char * s) {
-
+static void generate_integer_assign(struct generator * g, struct node * p, const char * s) {
+    write_comment(g, p);
     g->V[0] = p->name;
     g->S[0] = s;
     w(g, "~M~V0 ~S0 "); generate_AE(g, p->AE); w(g, "~N");
 }
 
-static void generate_integer_test(struct generator * g, struct node * p, char * s) {
-
-    w(g, "~Mif not ");
+static void generate_integer_test(struct generator * g, struct node * p) {
+    write_comment(g, p);
+    int relop = p->type;
+    int optimise_to_return = (g->failure_label == x_return && p->right && p->right->type == c_functionend);
+    if (optimise_to_return) {
+        w(g, "~Mreturn ");
+        p->right = NULL;
+    } else {
+        w(g, "~Mif ");
+        // We want the inverse of the snowball test here.
+        relop ^= 1;
+    }
     generate_AE(g, p->left);
-    write_char(g, ' ');
-    write_string(g, s);
-    write_char(g, ' ');
+    // Relational operators are the same as C.
+    write_c_relop(g, relop);
     generate_AE(g, p->AE);
-    w(g, ":");
-    write_block_start(g);
-    write_failure(g);
-    write_block_end(g);
-    g->unreachable = false;
+    if (optimise_to_return) {
+        w(g, "~N");
+    } else {
+        w(g, ":");
+        write_block_start(g);
+        write_failure(g);
+        write_block_end(g);
+        g->unreachable = false;
+    }
 }
 
 static void generate_call(struct generator * g, struct node * p) {
-
+    int signals = check_possible_signals_list(g, p->name->definition, c_define, 0);
     write_comment(g, p);
     g->V[0] = p->name;
-    write_failure_if(g, "not ~V0()", p);
+    if (g->failure_label == x_return &&
+        (signals == 0 || (p->right && p->right->type == c_functionend))) {
+        /* Always fails or tail call. */
+        writef(g, "~Mreturn ~V0()~N", p);
+        return;
+    }
+    if (signals == 1) {
+        /* Always succeeds. */
+        writef(g, "~M~V0()~N", p);
+    } else if (signals == 0) {
+        /* Always fails. */
+        writef(g, "~M~V0()~N", p);
+        write_failure(g);
+    } else {
+        write_failure_if(g, "not ~V0()", p);
+    }
 }
 
 static void generate_grouping(struct generator * g, struct node * p, int complement) {
+    write_comment(g, p);
 
-    struct grouping * q = p->name->grouping;
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->S[1] = complement ? "out" : "in";
     g->V[0] = p->name;
-    g->I[0] = q->smallest_ch;
-    g->I[1] = q->largest_ch;
-    write_failure_if(g, "not self.~S1_grouping~S0(~n.~W0, ~I0, ~I1)", p);
+    write_failure_if(g, "not self.~S1_grouping~S0(~n.~W0)", p);
 }
 
 static void generate_namedstring(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->V[0] = p->name;
@@ -1002,12 +1014,16 @@ static void generate_literalstring(struct generator * g, struct node * p) {
 
 static void generate_define(struct generator * g, struct node * p) {
     struct name * q = p->name;
+    if (q->type == t_routine && !q->used) return;
 
-    struct str * saved_output = g->outbuf;
+    write_newline(g);
+    write_comment(g, p);
 
     g->V[0] = q;
-    w(g, "~N~Mdef ~W0(self):~+~N");
+    w(g, "~Mdef ~W0(self):~+~N");
 
+    /* Save output. */
+    struct str * saved_output = g->outbuf;
     g->outbuf = str_new();
 
     g->next_label = 0;
@@ -1016,8 +1032,14 @@ static void generate_define(struct generator * g, struct node * p) {
     str_clear(g->failure_str);
     g->failure_label = x_return;
     g->unreachable = false;
+    int signals = check_possible_signals_list(g, p->left, c_define, 0);
     generate(g, p->left);
-    if (!g->unreachable) w(g, "~Mreturn True~N");
+    if (p->left->right) {
+        assert(p->left->right->type == c_functionend);
+        if (signals) {
+            generate(g, p->left->right);
+        }
+    }
     w(g, "~-");
 
     str_append(saved_output, g->outbuf);
@@ -1025,37 +1047,53 @@ static void generate_define(struct generator * g, struct node * p) {
     g->outbuf = saved_output;
 }
 
+static void generate_functionend(struct generator * g, struct node * p) {
+    (void)p;
+    w(g, "~Mreturn True~N");
+}
+
 static void generate_substring(struct generator * g, struct node * p) {
+    write_comment(g, p);
 
     struct among * x = p->among;
 
-    write_comment(g, p);
-
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->I[0] = x->number;
 
-    if (!x->amongvar_needed) {
-        write_failure_if(g, "self.find_among~S0(~n.a_~I0) == 0", p);
-    } else {
+    if (x->amongvar_needed) {
         writef(g, "~Mamong_var = self.find_among~S0(~n.a_~I0)~N", p);
-        write_failure_if(g, "among_var == 0", p);
+        if (!x->always_matches) {
+            write_failure_if(g, "among_var == 0", p);
+        }
+    } else if (x->always_matches) {
+        writef(g, "~Mself.find_among~S0(~n.a_~I0)~N", p);
+    } else {
+        write_failure_if(g, "self.find_among~S0(~n.a_~I0) == 0", p);
     }
 }
 
 static void generate_among(struct generator * g, struct node * p) {
-
     struct among * x = p->among;
 
-    if (x->substring == 0) generate_substring(g, p);
-
-    if (x->starter != 0) generate(g, x->starter);
+    if (x->substring == NULL) {
+        generate_substring(g, p);
+    } else {
+        write_comment(g, p);
+    }
 
     if (x->command_count == 1 && x->nocommand_count == 0) {
         /* Only one outcome ("no match" already handled). */
         generate(g, x->commands[0]);
     } else if (x->command_count > 0) {
-        int i;
-        for (i = 1; i <= x->command_count; i++) {
+        /* We dispatch the integer result in `among_var` with an if-chain,
+         * which is O(n) unless Python has a special optimisation (and
+         * profiling with the `timeit` module suggests it doesn't).  There
+         * doesn't appear to be a good alternative in Python (3.10 added
+         * `match` but that seems to be aimed more at pattern matching rather
+         * than O(1) dispatch of an integer and it was actually slower when we
+         * tried generating it here).
+         */
+        for (int i = 1; i <= x->command_count; i++) {
             if (i == x->command_count && x->nocommand_count == 0) {
                 w(g, "~Melse:~N~+");
             } else {
@@ -1071,20 +1109,17 @@ static void generate_among(struct generator * g, struct node * p) {
 }
 
 static void generate_booltest(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->V[0] = p->name;
     write_failure_if(g, "not ~V0", p);
 }
 
 static void generate_false(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     write_failure(g);
 }
 
 static void generate_debug(struct generator * g, struct node * p) {
-
     write_comment(g, p);
     g->I[0] = g->debug_count++;
     g->I[1] = p->line_number;
@@ -1092,14 +1127,10 @@ static void generate_debug(struct generator * g, struct node * p) {
 }
 
 static void generate(struct generator * g, struct node * p) {
-
-    int a0;
-    struct str * a1;
-
     if (g->unreachable) return;
 
-    a0 = g->failure_label;
-    a1 = str_copy(g->failure_str);
+    int a0 = g->failure_label;
+    struct str * a1 = str_copy(g->failure_str);
 
     switch (p->type) {
         case c_define:        generate_define(g, p); break;
@@ -1117,6 +1148,11 @@ static void generate(struct generator * g, struct node * p) {
         case c_do:            generate_do(g, p); break;
         case c_goto:          generate_GO(g, p, 1); break;
         case c_gopast:        generate_GO(g, p, 0); break;
+        case c_goto_grouping: generate_GO_grouping(g, p, 1, 0); break;
+        case c_gopast_grouping:
+                              generate_GO_grouping(g, p, 0, 0); break;
+        case c_goto_non:      generate_GO_grouping(g, p, 1, 1); break;
+        case c_gopast_non:    generate_GO_grouping(g, p, 0, 1); break;
         case c_repeat:        generate_repeat(g, p); break;
         case c_loop:          generate_loop(g, p); break;
         case c_atleast:       generate_atleast(g, p); break;
@@ -1154,12 +1190,14 @@ static void generate(struct generator * g, struct node * p) {
             generate_AE(g, p->AE);
             w(g, ")~N");
             break;
-        case c_eq:            generate_integer_test(g, p, "=="); break;
-        case c_ne:            generate_integer_test(g, p, "!="); break;
-        case c_gr:            generate_integer_test(g, p, ">"); break;
-        case c_ge:            generate_integer_test(g, p, ">="); break;
-        case c_ls:            generate_integer_test(g, p, "<"); break;
-        case c_le:            generate_integer_test(g, p, "<="); break;
+        case c_eq:
+        case c_ne:
+        case c_gt:
+        case c_ge:
+        case c_lt:
+        case c_le:
+            generate_integer_test(g, p);
+            break;
         case c_call:          generate_call(g, p); break;
         case c_grouping:      generate_grouping(g, p, false); break;
         case c_non:           generate_grouping(g, p, true); break;
@@ -1171,6 +1209,7 @@ static void generate(struct generator * g, struct node * p) {
         case c_false:         generate_false(g, p); break;
         case c_true:          break;
         case c_debug:         generate_debug(g, p); break;
+        case c_functionend:   generate_functionend(g, p); break;
         default: fprintf(stderr, "%d encountered\n", p->type);
                  exit(1);
     }
@@ -1181,7 +1220,6 @@ static void generate(struct generator * g, struct node * p) {
 }
 
 static void generate_class_begin(struct generator * g) {
-
     w(g, "from .basestemmer import ");
     w(g, g->options->parent_class_name);
     w(g, "~N"
@@ -1201,82 +1239,60 @@ static void generate_class_begin(struct generator * g) {
 }
 
 static void generate_among_table(struct generator * g, struct among * x) {
+    write_newline(g);
+    write_comment(g, x->node);
 
     struct amongvec * v = x->b;
 
     g->I[0] = x->number;
 
     w(g, "~Ma_~I0 = [~N~+");
-    {
-        int i;
-        for (i = 0; i < x->literalstring_count; i++) {
-            g->I[0] = v->i;
-            g->I[1] = v->result;
-            g->L[0] = v->b;
-            g->S[0] = i < x->literalstring_count - 1 ? "," : "";
+    for (int i = 0; i < x->literalstring_count; i++) {
+        g->I[0] = v[i].i;
+        g->I[1] = v[i].result;
+        g->L[0] = v[i].b;
+        g->S[0] = i < x->literalstring_count - 1 ? "," : "";
 
-            w(g, "~MAmong(~L0, ~I0, ~I1");
-            if (v->function != 0) {
-                w(g, ", \"");
-                if (v->function->type == t_routine) {
-                    /* Need to use mangled version of private name here. */
-                    w(g, "_~n");
-                }
-                write_varname(g, v->function);
-                w(g, "\"");
-            }
-            w(g, ")~S0~N");
-            v++;
+        w(g, "~MAmong(~L0, ~I0, ~I1");
+        if (v[i].function != NULL) {
+            w(g, ", ");
+            write_varname(g, v[i].function);
         }
+        w(g, ")~S0~N");
     }
-    w(g, "~-~M]~N~N");
+    w(g, "~-~M]~N");
 }
 
 static void generate_amongs(struct generator * g) {
-    struct among * x;
-    for (x = g->analyser->amongs; x; x = x->next) {
+    for (struct among * x = g->analyser->amongs; x; x = x->next) {
         generate_among_table(g, x);
     }
 }
 
-static void set_bit(symbol * b, int i) { b[i/8] |= 1 << i%8; }
-
 static void generate_grouping_table(struct generator * g, struct grouping * q) {
-
-    int range = q->largest_ch - q->smallest_ch + 1;
-    int size = (range + 7)/ 8;  /* assume 8 bits per symbol */
     symbol * b = q->b;
-    symbol * map = create_b(size);
-    int i;
-    for (i = 0; i < size; i++) map[i] = 0;
-
-    /* Using unicode would require revision here */
-
-    for (i = 0; i < SIZE(b); i++) set_bit(map, b[i] - q->smallest_ch);
 
     g->V[0] = q->name;
 
-    w(g, "~M~W0 = [");
-    for (i = 0; i < size; i++) {
-        write_int(g, map[i]);
-        if (i < size - 1) w(g, ", ");
+    // We could use frozenset, but it seems slightly slower to construct which
+    // adds to startup time.
+    w(g, "~M~W0 = {");
+    for (int i = 0; i < SIZE(b); i++) {
+        if (i > 0) w(g, ", ");
+        write_literal_char(g, b[i]);
     }
-    w(g, "]~N~N");
-    lose_b(map);
+    w(g, "}~N~N");
 }
 
 static void generate_groupings(struct generator * g) {
-    struct grouping * q;
-    for (q = g->analyser->groupings; q; q = q->next) {
+    for (struct grouping * q = g->analyser->groupings; q; q = q->next) {
         if (q->name->used)
             generate_grouping_table(g, q);
     }
 }
 
 static void generate_members(struct generator * g) {
-
-    struct name * q;
-    for (q = g->analyser->names; q; q = q->next) {
+    for (struct name * q = g->analyser->names; q; q = q->next) {
         g->V[0] = q;
         switch (q->type) {
             case t_string:
@@ -1293,9 +1309,8 @@ static void generate_members(struct generator * g) {
 }
 
 static void generate_methods(struct generator * g) {
-
     struct node * p = g->analyser->program;
-    while (p != 0) {
+    while (p != NULL) {
         generate(g, p);
         g->unreachable = false;
         p = p->right;
@@ -1304,18 +1319,19 @@ static void generate_methods(struct generator * g) {
 
 static void generate_label_classes(struct generator * g)
 {
-    int i;
-    for (i = 0; i <= g->max_label; i++) {
+    for (int i = 0; i <= g->max_label; i++) {
         g->I[0] = i;
         w(g, "~N~Nclass lab~I0(BaseException): pass~N");
     }
 }
 
 extern void generate_program_python(struct generator * g) {
-
     g->outbuf = str_new();
     g->failure_str = str_new();
 
+    // Only needed for Python 2, which defaults to ASCII.
+    w(g, "#-*- coding: utf-8 -*-~N");
+
     write_start_comment(g, "# ", NULL);
     if (g->analyser->int_limits_used) {
         /* sys.maxsize is used in the code generated for maxint and minint */
@@ -1323,12 +1339,13 @@ extern void generate_program_python(struct generator * g) {
     }
     generate_class_begin(g);
 
-    generate_amongs(g);
     generate_groupings(g);
 
     generate_members(g);
     generate_methods(g);
 
+    generate_amongs(g);
+
     generate_label_classes(g);
 
     output_str(g->options->output_src, g->outbuf);
